from urllib.parse import quote


class RequestUtilities:
    """Requests utilities."""

    @staticmethod
    def quote(
        uri: str,
        *args
    ) -> str:
        """Url Encode.

        Args:
            uri (str): uri to encode
            args: urllib.parse.quote original args

        Returns:
            str: encoded uri
        """  # noqa: DAR101
        return quote(uri, *args).replace('%24', '$').replace('%27', "'")



from __future__ import annotations
from typing import TYPE_CHECKING
from lime_python import Command
from ...extension_base import ExtensionBase
from .content_type import ContentType
from .uri_templates import UriTemplates

if TYPE_CHECKING:
    from ....client import Client

POSTMASTER_AI = 'postmaster@ai'


class AiModelExtension(ExtensionBase):
    """Extension to handle Blip Analytics Services."""

    def __init__(self, client: Client, domain: str) -> None:
        super().__init__(client, f'{POSTMASTER_AI}.{domain}')

    # Model

    async def get_models_async(
        self,
        skip: int = 0,
        take: int = 100,
        ascending: bool = False,
        **kwargs
    ) -> Command:
        """Search in all trained and/or published models.

        Args:
            skip (int): Number of models to be skipped.
            take (int): Number of model to be take.
            ascending (bool): Sets ascending alphabetical order.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: Command response
        """
        models_resource_query = self.build_resource_query(
            UriTemplates.MODELS,
            {
                '$skip': skip,
                '$take': take,
                '$ascending': ascending,
                **kwargs
            }
        )

        return await self.process_command_async(
            self.create_get_command(
                models_resource_query,
            )
        )

    async def get_model_async(self, id: str) -> Command:
        """Get specific AI model.

        Args:
            id (str): Model id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_get_command(
                self.build_uri(UriTemplates.MODEL, id)
            )
        )

    async def get_model_summary_async(self) -> Command:
        """Get model summary.

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_get_command(UriTemplates.MODELS_SUMMARY)
        )

    async def get_last_trained_or_published_model_async(self) -> Command:
        """Get last trained or published model.

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_get_command(UriTemplates.LAST_TRAINED_OR_PUBLISH_MODEL)
        )

    async def train_model_async(self) -> Command:
        """Train model.

        Returns:
            Command: Command response
        """
        train_model_command = self.create_set_command(
            UriTemplates.MODELS,
            {},
            ContentType.MODEL_TRAINING
        )

        return await self.process_command_async(train_model_command)

    async def publish_model_async(self, id: str) -> Command:
        """Publish an existing artificial intelligence model.

        Args:
            id (str): model id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_set_command(
                UriTemplates.MODELS,
                {
                    'id': id
                },
                ContentType.MODEL_PUBLISHING
            )
        )



from dataclasses import dataclass, field
from typing import Dict
from uuid import uuid4

from lime_python import (Authentication, GuestAuthentication,
                         SessionCompression, SessionEncryption)


@dataclass
class Application:
    """Basic Application with default values."""

    identifier: str = str(uuid4())
    compression: str = SessionCompression.NONE
    encryption: str = SessionEncryption.NONE
    instance: str = 'default'
    domain: str = 'msging.net'
    scheme: str = 'wss'
    hostname: str = 'ws.msging.net'
    port: int = 443
    presence: Dict[str, str] = field(
        default_factory=lambda: {
            'status': 'available',
            'routingRule': 'identity'
        }
    )
    notify_consumed: bool = True
    authentication: Authentication = GuestAuthentication()
    command_timeout: int = 6  # in seconds



class ContentType:
    """Word set content types."""

    WORD_SET = 'application/vnd.iris.ai.word-set+json'

    WORD_SET_ANALYSIS = 'application/vnd.iris.ai.word-set-analysis+json'

    WORD_SET_WORD = 'application/vnd.iris.ai.word+json'



class ContentType:
    """AI Content type class."""

    ANALYSIS = 'application/vnd.iris.ai.analysis-request+json'

    ANALYSIS_FEEDBACK = 'application/vnd.iris.ai.analysis-feedback+json'

    JSON_DOCUMENT = 'application/json'

    CONFUSION_MATRIX = 'application/vnd.iris.ai.confusion-matrix+json'




from __future__ import annotations
from typing import TYPE_CHECKING, List

from lime_python import Command, ContentTypes

from ...extension_base import ExtensionBase
from .content_type import ContentType
from .uri_templates import UriTemplates

if TYPE_CHECKING:
    from ....client import Client

POSTMASTER_AI = 'postmaster@ai'


class AnalyticsExtension(ExtensionBase):
    """Extension to handle Blip Analytics Services."""

    def __init__(self, client: Client, domain: str) -> None:
        super().__init__(client, f'{POSTMASTER_AI}.{domain}')

    async def get_analysis_async(
        self,
        skip: int = 0,
        take: int = 100,
        ascending: bool = False,
        filter: str = None,
        intents: List[str] = None,
        feedbacks: list = None,
        source: str = None,
        begin_date: str = None,
        end_date: str = None,
        min_score: str = None,
        max_score: str = None,
        **kwargs
    ) -> Command:
        """Retrieve the history of performed analysis.

        Args:
            skip (int): The number of elements to be skipped.
            take (int): Limit of total of items to be returned.
            ascending (bool): Sets ascending alphabetical order.
            filter (str): request filter options.
            intents (List[str]): List of intents.
            feedbacks (list): the analysis feedback.
            source (str): the source provider.
            begin_date (str): the begin date of the request.
            end_date (str): the end date of the request.
            min_score (str): minimal score.
            max_score (str): maximus score.
            kwargs (any): key args.

        Returns:
            Command: Command response
        """
        analysis_params = {
            '$skip': skip,
            '$take': take,
            '$ascending': ascending,
            '$filter': filter,
            'intents': intents,
            'feedbaccks': feedbacks,
            'source': source,
            'beginDate': begin_date,
            'endDate': end_date,
            'minScore': min_score,
            'maxScore': max_score,
            **kwargs
        }

        uri = self.build_resource_query(UriTemplates.ANALYSIS, analysis_params)
        get_command = self.create_get_command(uri)

        return await self.process_command_async(get_command)

    async def analyse_async(self, analysis: dict) -> Command:
        """Analyzes an user sentence using a published model.

        Args:
            analysis (dict): Input analysis

        Returns:
            Command: Command response
        """
        analyse_command = self.create_set_command(
            UriTemplates.ANALYSIS,
            type_n=ContentType.ANALYSIS,
            resource=analysis
        )

        return await self.process_command_async(analyse_command)

    async def set_analysis_by_email_async(
        self,
        email: str,
        filter: str,
        intents: list = None,
        feedbacks: list = None,
        source: str = None,
        begin_date: str = None,
        end_date: str = None,
        min_score: str = None,
        max_score: str = None,
        **kwargs
    ) -> Command:
        """Send analysis by email.

        Args:
            email (str): email to be sended
            filter (str): filter OData
            intents (list): list of intents.
            feedbacks (list): feedback list.
            source (str): analysis source.
            begin_date (str): the begin date of the request.
            end_date (str): the end date of the request.
            min_score (str): min score to be considered.
            max_score (str): max score to be considered.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: Command response
        """
        send_email_resource = {
            'email': email,
            'filter': self.build_resource_query(
                UriTemplates.ANALYSIS,
                {
                    '$filter': filter,
                    'intents': intents,
                    'feedbacks': feedbacks,
                    'source': source,
                    'beginDate': begin_date,
                    'endDate': end_date,
                    'minScore': min_score,
                    'maxScore': max_score,
                    **kwargs
                }
            )
        }
        send_email_command = self.create_set_command(
            UriTemplates.ANALYSIS_EMAIL,
            send_email_resource,
            ContentType.JSON_DOCUMENT
        )

        return await self.process_command_async(send_email_command)

    async def set_analysis_feedback_async(
        self,
        id: str,
        analyses: list
    ) -> Command:
        """Send feedbacks into analysis.

        Args:
            id (str): Command id
            analyses (list): analyses

        Returns:
            Command: Command response
        """
        analyses_feedback_command = self.create_set_command(
            self.build_uri(UriTemplates.ANALYSES_FEEDBACK, id),
            analyses,
            ContentType.ANALYSIS_FEEDBACK
        )
        return await self.process_command_async(analyses_feedback_command)

    async def set_analyses_feedback_async(
        self,
        id: str,
        analyses: list
    ) -> Command:
        """Send feedbacks into analysis.

        Args:
            id (str): Command id
            analyses (list): analyses

        Returns:
            Command: Command response
        """
        analyses_feedback_resource = {
            'itemType': ContentType.ANALYSIS_FEEDBACK,
            'items': analyses
        }
        analyses_feedback_command = self.create_set_command(
            self.build_uri(UriTemplates.ANALYSES_FEEDBACK, id),
            analyses_feedback_resource,
            ContentTypes.COLLECTION
        )
        return await self.process_command_async(analyses_feedback_command)

    async def get_analytics_async(self, id: str = None) -> Command:
        """Get analytics.

        Args:
            id (str): Unique identifier of the command.

        Returns:
            Command: Command response
        """
        uri = self.build_uri(
            UriTemplates.ANALYTICS_ID,
            id
        ) if id else UriTemplates.ANALYTICS

        return await self.process_command_async(self.create_get_command(uri))

    async def set_analytics_async(self, confusion_matrix: dict) -> Command:
        """Create a confusion matrix into your model.

        Args:
            confusion_matrix (dict): Represents a confusion matrix model.

        Returns:
            Command: Command response
        """
        confusion_matrix_resource = self.create_set_command(
            UriTemplates.ANALYTICS,
            confusion_matrix,
            ContentType.CONFUSION_MATRIX
        )

        return await self.process_command_async(confusion_matrix_resource)

    async def delete_analytics_async(self, id: str) -> Command:
        """Delete analytics.

        Args:
            id (str): Analytics idd

        Returns:
            Command: Command response
        """
        delete_analytics_command = self.create_delete_command(
            self.build_uri(UriTemplates.ANALYTICS_ID, id)
        )

        return await self.process_command_async(delete_analytics_command)



from typing import Any


class ClassUtilities:
    """Classes utilities."""

    @staticmethod
    def merge_dataclasses(*args) -> Any:
        """Merge two or more dataclasses.

        Adapted solution from:
        https://stackoverflow.com/questions/9667818/python-how-to-merge-two-class

        Args:
            args: dataclasses to merge

        Raises:
            ValueError: dataclasses of different types were passed

        Returns:
            Any: the merged dataclass
        """
        if len({arg.__class__.__name__ for arg in args}) > 1:
            raise ValueError(
                'Merge of non-homogeneous entries no allowed.'
            )
        data = {}
        for entry in args:
            data.update(
                ClassUtilities.__remove_none_values(
                    vars(entry)  # noqa: WPS421
                )
            )

        return args[-1].__class__(**data)

    @staticmethod
    def __remove_none_values(data: dict) -> dict:
        clear_data: dict = {}
        for key, value in data.items():
            if value is not None:
                clear_data[key] = value
        return clear_data



from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .ai_model import AiModelExtension
from .analytics import AnalyticsExtension
from .content_assistant import ContentAssistantExtension
from .entities import EntitiesExtension
from .intents import IntentsExtension
from .word_set import WordSetExtension

if TYPE_CHECKING:
    from ...client import Client


@dataclass
class AiExtension:
    """AI bundled extensions."""

    client: Client
    to: str
    ai_model: AiModelExtension = field(init=False)
    analytics: AnalyticsExtension = field(init=False)
    content_assistant: ContentAssistantExtension = field(init=False)
    entities: EntitiesExtension = field(init=False)
    intents: IntentsExtension = field(init=False)
    word_set: WordSetExtension = field(init=False)

    def __post_init__(self) -> None:  # noqa: D105
        self.ai_model = AiModelExtension(self.client, self.to)
        self.analytics = AnalyticsExtension(self.client, self.to)
        self.content_assistant = ContentAssistantExtension(
            self.client,
            self.to
        )
        self.entities = EntitiesExtension(self.client, self.to)
        self.intents = IntentsExtension(self.client, self.to)
        self.word_set = WordSetExtension(self.client, self.to)



class UriTemplates:
    """Blip AI URI templates."""

    ANALYSIS = '/analysis'

    ANALYSIS_FEEDBACK = '/analysis/{0}/feedback'

    ANALYSES_FEEDBACK = '/analysis/feedback'

    ANALYSIS_EMAIL = '/enhancement/send-by-email'

    ANALYTICS = '/analytics/confusion-matrix'

    ANALYTICS_ID = '/analytics/confusion-matrix/{0}'



from __future__ import annotations
from typing import TYPE_CHECKING
from lime_python import Command
from ...extension_base import ExtensionBase
from .content_type import ContentType
from .uri_templates import UriTemplates

if TYPE_CHECKING:
    from ....client import Client

POSTMASTER_AI = 'postmaster@ai'


class EntitiesExtension(ExtensionBase):
    """Extension to handle Blip Entities Services."""

    def __init__(self, client: Client, domain: str) -> None:
        super().__init__(client, f'{POSTMASTER_AI}.{domain}')

    async def get_entity_async(self, id: str) -> Command:
        """Get entity.

        Args:
            id (str): entity id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_get_command(
                self.build_uri(UriTemplates.ENTITY, id)
            )
        )

    async def get_entities_async(
        self,
        skip: int = 0,
        take: int = 100,
        ascending: bool = False,
        name: str = None,
        **kwargs
    ) -> Command:
        """Get all entities.

        Args:
            skip (int): Numbers of entities to be skipped.
            take (int): Numbers of entities to be taked.
            ascending (bool): Sets ascending alphabetical order..
            name (str): name.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: Command response
        """
        entities_resource_query = self.build_resource_query(
            UriTemplates.ENTITIES,
            {
                '$skip': skip,
                '$take': take,
                '$ascending': ascending,
                'name': name,
                **kwargs
            }
        )

        return await self.process_command_async(
            self.create_get_command(
                entities_resource_query
            )
        )

    async def set_entity_async(self, entity: dict) -> Command:
        """Add entity on a base.

        Args:
            entity (dict): Entity to be added

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_set_command(
                UriTemplates.ENTITIES,
                entity,
                ContentType.ENTITY
            )
        )

    async def delete_entity_async(self, id: str) -> Command:
        """Delete a entity.

        Args:
            id (str): Entity id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_delete_command(
                self.build_uri(UriTemplates.ENTITY, id)
            )
        )

    async def delete_entities_async(self) -> Command:
        """Delete all entities.

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_delete_command(UriTemplates.ENTITIES)
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from lime_python import Command
from ..extension_base import ExtensionBase
from .uri_templates import UriTemplates

if TYPE_CHECKING:
    from ...client import Client

POSTMASTER_MEDIA = 'postmaster@media'


class MediaExtension(ExtensionBase):
    """Extension to handle blip media."""

    def __init__(self, client: Client, domain: str) -> None:
        super().__init__(client, f'{POSTMASTER_MEDIA}.{domain}')

    async def get_upload_token_async(
        self,
        secure: bool = None,
        **kwargs
    ) -> Command:
        """Get token to upload media.

        Args:
            secure (bool): is media secure.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command
        """
        uri = self.build_resource_query(
            UriTemplates.MEDIA_UPLOAD,
            {'secure': secure, **kwargs}
        )
        return await self.process_command_async(self.create_get_command(uri))

    async def refresh_media_async(
        self,
        id: str,
        **kwargs
    ) -> Command:
        """Refresh an uploaded media.

        Args:
            id (str): the media id
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command
        """
        uri = self.build_uri(UriTemplates.REFRESH_MEDIA, id)
        uri = self.build_resource_query(uri, kwargs)
        return await self.process_command_async(self.create_get_command(uri))



from typing import Callable
from uuid import uuid4

from lime_python import Envelope


class Receiver:
    """Receiver base class."""

    def __init__(
        self,
        predicate: Callable[[Envelope], bool],
        callback: Callable[[Envelope], None]
    ) -> None:
        self.predicate: Callable[[Envelope], bool] = predicate
        self.callback: Callable[[Envelope], None] = callback
        self.id = str(uuid4())

    def __eq__(self, other: object) -> bool:  # noqa: D105
        if isinstance(other, Receiver):
            return self.id == other.id
        return False

    def __hash__(self) -> int:  # noqa: D105
        return hash(self.id)

    @property
    def predicate(self) -> Callable[[Envelope], bool]:  # noqa: D102
        return self.__predicate

    @predicate.setter
    def predicate(self, value: Callable[[Envelope], bool]) -> None:
        if not isinstance(value, Callable):
            self.__predicate = lambda _: value in [True, None]  # noqa: WPS510
            return
        self.__predicate = value



from typing import Any

from lime_python import Command
from ..extension_base import ExtensionBase
from .content_types import ContentTypes
from .uri_templates import UriTemplates


class ChatExtension(ExtensionBase):
    """Extension to handle blip chat history."""

    async def get_threads_async(
        self,
        take: int = None,
        skip: int = None,
        message_date: str = None,
        refresh_expired_media: bool = None,
        **kwargs
    ) -> Command:
        """Get chatbot message history.

        Args:
            take (int):	Limit of total of items to be returned. \
            The maximum value allowed is 100
            skip (int): The number of elements to be skipped
            message_date (str): Initial date on the threads query
            refresh_expired_media (bool): Defines if the expired media \
            links should be refreshed
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command
        """
        uri = self.build_resource_query(
            UriTemplates.THREADS,
            {
                '$take': take,
                '$skip': skip,
                'messageDate': message_date,
                'refreshExpiredMedia': refresh_expired_media,
                **kwargs
            }
        )
        return await self.process_command_async(self.create_get_command(uri))

    async def get_thread_async(
        self,
        identity: str = None,
        take: int = None,
        skip: int = None,
        message_id: str = None,
        storage_date: str = None,
        direction: str = None,
        refresh_expired_media: bool = None,
        decrypt_content: bool = None,
        after: str = None,
        **kwargs
    ) -> Command:
        """Get a user thread message history.

        Args:
            identity (str): user identity
            take (int):	Limit of total of items to be returned. \
            The maximum value allowed is 100
            skip (int): The number of elements to be skipped
            message_id (str): Initial message id for the thread messages query
            storage_date (str): The reference date to search. \
            Example: 2020-01-08T15:59:07.086Z
            direction (str): Possible values: asc and desc. \
            Define whether messages will be returned \
            after(in ascending order) \
            or before(in descending order) a date, respectively. \
            Needs storageDate or messageId to be defined
            refresh_expired_media (bool): Defines if the expired \
            media links should be refreshed
            decrypt_content (bool): decrypt the content
            after (str): after
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: [description]
        """
        uri = self.build_uri(UriTemplates.THREAD, identity)
        uri = self.build_resource_query(
            uri,
            {
                '$take': take,
                '$skip': skip,
                'messageId': message_id,
                'storageDate': storage_date,
                'direction': direction,
                'refreshExpiredMedia': refresh_expired_media,
                'decryptContent': decrypt_content,
                'after': after,
                **kwargs
            }
        )
        return await self.process_command_async(self.create_get_command(uri))

    async def get_thread_unread_messages_async(
        self,
        identity: str,
        **kwargs
    ) -> Command:
        """Get user unread messages.

        Args:
            identity (str): user identity
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command
        """
        uri = self.build_uri(UriTemplates.THREAD_UNREAD_MESSAGES, identity)
        uri = self.build_resource_query(uri, kwargs)
        return await self.process_command_async(self.create_get_command(uri))

    async def set_thread_async(
        self,
        identity: str,
        thread: Any,
        **kwargs
    ) -> Command:
        """Set a user thread.

        Args:
            identity (str): user identity
            thread (Any): message thread
            kwargs: any other optional parameter not covered by the methods

        Returns:
            Command
        """
        uri = self.build_uri(UriTemplates.THREAD, identity)
        uri = self.build_resource_query(uri, kwargs)
        return await self.process_command_async(
            self.create_set_command(uri, thread, ContentTypes.THREAD_MESSAGE)
        )



from __future__ import annotations
from typing import TYPE_CHECKING
from lime_python import Command, ContentTypes
from ...extension_base import ExtensionBase
from .content_type import ContentType
from .uri_templates import UriTemplates

if TYPE_CHECKING:
    from ....client import Client

POSTMASTER_AI = 'postmaster@ai'


class WordSetExtension(ExtensionBase):
    """Extension to handle Worsd Set AI Services."""

    def __init__(self, client: Client, domain: str) -> None:
        super().__init__(client, f'{POSTMASTER_AI}.{domain}')

    async def get_word_set_async(
        self,
        id: str,
        deep: bool = False,
        **kwargs
    ) -> Command:
        """Get word set.

        Args:
            id (str): word set id
            deep (bool): deep.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: Command response
        """
        word_set_resource_query = self.build_resource_query(
            self.build_uri(UriTemplates.WORD_SET, id),
            {
                'deep': deep,
                **kwargs
            }
        )

        return await self.process_command_async(
            self.create_get_command(
                word_set_resource_query
            )
        )

    async def set_word_set_resource_async(
        self,
        id: str,
        resource: dict
    ) -> Command:
        """Create word set.

        Args:
            id (str): word set id
            resource (dict): word set resource

        Returns:
            Command: Command response
        """
        set_resource_command = self.create_set_command(
            self.build_uri(UriTemplates.WORD_SET, id),
            type_n=ContentTypes.COLLECTION,
            resource={
                'itemType': ContentType.WORD_SET_WORD,
                'items': resource
            }
        )

        return await self.process_command_async(set_resource_command)

    async def set_word_set_async(self, word_set: dict) -> Command:
        """Create word set.

        Args:
            word_set (dict): word set object

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_set_command(
                UriTemplates.WORD_SETS,
                word_set,
                ContentType.WORD_SET
            )
        )

    async def delete_word_set_async(self, id: str) -> Command:
        """Delete word set.

        Args:
            id (str): word set id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_delete_command(
                self.build_uri(UriTemplates.WORD_SET, id)
            )
        )

    async def analyse_word_set_async(self, analysis: dict) -> Command:
        """Analyse word set.

        Args:
            analysis (dict): Analysis

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_set_command(
                UriTemplates.WORD_SETS_ANALYSIS,
                analysis,
                ContentType.WORD_SET_ANALYSIS
            )
        )



from __future__ import annotations
from typing import TYPE_CHECKING
from lime_python import Command, ContentTypes
from ...extension_base import ExtensionBase
from .content_type import ContentType
from .uri_templates import UriTemplates

if TYPE_CHECKING:
    from ....client import Client

POSTMASTER_AI = 'postmaster@ai'


class ContentAssistantExtension(ExtensionBase):
    """Extension to handle Content Assistant Services."""

    def __init__(self, client: Client, domain: str) -> None:
        super().__init__(client, f'{POSTMASTER_AI}.{domain}')

    async def analyse_content_async(self, analysis: dict) -> Command:
        """Analyse content.

        Args:
            analysis (dict): Content Analysis object

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_set_command(
                UriTemplates.CONTENT_ANALYSIS,
                analysis,
                ContentType.ANALYSIS
            )
        )

    async def match_content_async(self, combination: dict) -> Command:
        """Match content.

        Args:
            combination (dict): combination object

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_set_command(
                UriTemplates.CONTENT_ANALYSIS,
                combination,
                ContentType.CONTENT_COMBINATION
            )
        )

    async def get_contents_async(
        self,
        skip: int = 0,
        take: int = 100,
        ascending: bool = False,
        intents: str = None,
        entities: str = None,
        text: str = None,
        begin_date: str = None,
        end_date: str = None,
        **kwargs
    ) -> Command:
        """Get contents.

        Args:
            skip (int): Number of contents to be skipped.
            take (int): Number of contents to be take.
            ascending (bool): Sets ascending alphabetical order..
            intents (str): Intents list.
            entities (str): Entities list.
            text (str): text.
            begin_date (str): begin date of contents.
            end_date (str): end date of contents.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: Command response
        """
        contents_resource_query = self.build_resource_query(
            UriTemplates.CONTENT,
            {
                '$skip': skip,
                '$take': take,
                '$ascending': ascending,
                'intents': intents,
                'entities': entities,
                'text': text,
                'beginDate': begin_date,
                'endDate': end_date,
                **kwargs
            }
        )

        return await self.process_command_async(
            self.create_get_command(contents_resource_query)
        )

    async def get_content_async(self, id: str) -> Command:
        """Get content.

        Args:
            id (str): Content id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_get_command(
                self.build_uri(UriTemplates.CONTENT_ID, id)
            )
        )

    async def set_content_async(self, content: dict) -> Command:
        """Create content combination.

        Args:
            content (dict): Content object

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_set_command(
                UriTemplates.CONTENT,
                content,
                ContentType.CONTENT_RESULT
            )
        )

    async def set_content_result_async(
        self,
        id: str,
        content: dict
    ) -> Command:
        """Set content result.

        Args:
            id (str): Content id.
            content (dict): Content object.

        Returns:
            Command: Command response
        """
        content_result_command = self.create_set_command(
            self.build_uri(UriTemplates.CONTENT_ID, id),
            content,
            ContentType.CONTENT_RESULT
        )

        return await self.process_command_async(content_result_command)

    async def set_content_combination_async(
        self,
        id: str,
        combination: dict
    ) -> Command:
        """Set content combination.

        Args:
            id (str): Combination id
            combination (dict): Combination object

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_set_command(
                self.build_uri(UriTemplates.CONTENT_ID, id),
                combination,
                ContentType.CONTENT_COMBINATION
            )
        )

    async def set_content_combinations_async(
        self,
        id: str,
        combinations=list
    ) -> Command:
        """Set list of combinations.

        Args:
            id (str): Combinations id.
            combinations (list, optional): Combinations list.

        Returns:
            Command: Command response
        """
        combinations_command = self.create_set_command(
            self.build_uri(UriTemplates.CONTENT_ID, id),
            type_n=ContentTypes.COLLECTION,
            resource={
                'itemType': ContentType.CONTENT_COMBINATION,
                'items': combinations
            }
        )

        return await self.process_command_async(combinations_command)

    async def delete_content_async(self, id: str) -> Command:
        """Delete content.

        Args:
            id (str): Content id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_delete_command(
                self.build_uri(UriTemplates.CONTENT_ID, id)
            )
        )



class UriTemplates:
    """Entities uri templates."""

    INTENTIONS = '/intentions'

    INTENTION = '/intentions/{0}'

    INTENTION_ANSWERS = '/intentions/{0}/answers'

    INTENTION_ANSWER = '/intentions/{0}/answers/{1}'

    INTENTION_QUESTIONS = '/intentions/{0}/questions'

    INTENTION_QUESTION = '/intentions/{0}/questions/{1}'



from __future__ import annotations
from typing import TYPE_CHECKING
from lime_python import Command, ContentTypes
from ...extension_base import ExtensionBase
from .content_type import ContentType
from .uri_templates import UriTemplates

if TYPE_CHECKING:
    from ....client import Client

POSTMASTER_AI = 'postmaster@ai'


class IntentsExtension(ExtensionBase):
    """Extension to handle Blip AI Services."""

    def __init__(self, client: Client, domain: str) -> None:
        super().__init__(client, f'{POSTMASTER_AI}.{domain}')

    async def get_intent_async(
        self,
        id: str,
        deep: bool = False,
        **kwargs
    ) -> Command:
        """Get a specific intent.

        Args:
            id (str): Unique identifier of the command.
            deep (bool): deep.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: Command response
        """
        get_intent_command = self.create_get_command(
            self.build_resource_query(
                self.build_uri(UriTemplates.INTENTION, id),
                {
                    'deep': deep,
                    **kwargs
                }
            )
        )

        return await self.process_command_async(get_intent_command)

    async def get_intents_async(
        self,
        skip: int = 0,
        take: int = 100,
        deep: bool = False,
        name: str = None,
        ascending: bool = False,
        **kwargs
    ) -> Command:
        """Get all intents from a model.

        Args:
            skip (int): The number of intents to be skipped.
            take (int): The number of intents to be returned.
            deep (bool): deep.
            name (str): intent name.
            ascending (bool): Sets ascending alphabetical order.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: Command response
        """
        intents_resource_query = self.build_resource_query(
            UriTemplates.INTENTIONS,
            {
                '$skip': skip,
                '$take': take,
                'deep': deep,
                'name': name,
                '$ascending': ascending,
                **kwargs
            }
        )

        return await self.process_command_async(
            self.create_get_command(intents_resource_query)
        )

    async def set_intent_async(self, intent: dict) -> Command:
        """Create intent.

        Args:
            intent (dict): intent object

        Returns:
            Command: Response Command
        """
        return await self.process_command_async(
            self.create_set_command(
                UriTemplates.INTENTIONS,
                intent,
                ContentType.INTENTION
            )
        )

    async def set_intents_async(self, intents: list) -> Command:
        """Create list of intents.

        Args:
            intents (list): List of intents

        Returns:
            Command: Response command
        """
        set_intents_resource = self.__build_resource(
            intents,
            ContentType.INTENTION
        )

        set_intents_command = self.create_set_command(
            UriTemplates.INTENTIONS,
            set_intents_resource,
            ContentTypes.COLLECTION
        )

        return await self.process_command_async(set_intents_command)

    async def merge_intent_async(self, intent: dict) -> Command:
        """Merge an intent into a base.

        Args:
            intent (dict): Intent to be merged

        Returns:
            Command: Response command
        """
        merge_intents_command = self.create_merge_command(
            UriTemplates.INTENTIONS,
            intent,
            ContentType.INTENTION
        )

        return await self.process_command_async(merge_intents_command)

    async def merge_intents_async(self, intents: list) -> Command:
        """Merge list of intents into a base.

        Args:
            intents (list): intents list

        Returns:
            Command: Command response
        """
        merge_intents_resource = self.__build_resource(
            intents,
            ContentType.INTENTION
        )

        merge_intents_command = self.create_merge_command(
            UriTemplates.INTENTIONS,
            merge_intents_resource,
            ContentTypes.COLLECTION
        )

        return await self.process_command_async(merge_intents_command)

    async def delete_intent_async(self, id: str) -> Command:
        """Delete intent from base.

        Args:
            id (str): Intent identifier

        Returns:
            Command: Command response
        """
        delete_intent_uri = self.build_uri(UriTemplates.INTENTION, id)

        return await self.process_command_async(
            self.create_delete_command(delete_intent_uri)
        )

    async def delete_intents_async(self):
        """Delete all intents from base.

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_delete_command(UriTemplates.INTENTIONS)
        )

    # Intent Answers

    async def get_intent_answers_async(
        self,
        id: str,
        skip: int = 0,
        take: int = 100,
        ascending: bool = False,
        **kwargs
    ) -> Command:
        """Get intent answers.

        Args:
            id (str): Intent id
            skip (int): Number of answers to be skipped.
            take (int): Number of answers to be take.
            ascending (bool): Sets ascending alphabetical order.
            kwargs: any other optional parameter not covered by the method

        Returns:
            Command: Command response
        """
        intent_answers_query = self.build_resource_query(
            self.build_uri(UriTemplates.INTENTION_ANSWERS, id),
            {
                '$skip': skip,
                '$take': take,
                '$ascending': ascending,
                **kwargs
            }
        )

        return await self.process_command_async(
            self.create_get_command(
                intent_answers_query
            )
        )

    async def set_intent_answers_async(
        self,
        id: str,
        answers: list
    ) -> Command:
        """Set a intent answer.

        Args:
            id (str): Intent id
            answers (list): Answers list

        Returns:
            Command: Command response
        """
        intent_answers_command = self.create_set_command(
            self.build_uri(UriTemplates.INTENTION_ANSWERS, id),
            type_n=ContentTypes.COLLECTION,
            resource=self.__build_resource(
                answers,
                ContentType.ANSWER
            )
        )

        return await self.process_command_async(intent_answers_command)

    async def delete_intent_answers_async(
        self,
        id: str,
        answer_id: str
    ) -> Command:
        """Delete intent answer.

        Args:
            id (str): intent id.
            answer_id (str): answer id.

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_delete_command(
                self.build_uri(
                    UriTemplates.INTENTION_ANSWER,
                    id,
                    answer_id
                )
            )
        )

    # Intent Questions
    async def get_intent_questions_async(self, id: str):
        """Get intent questions.

        Args:
            id (str): Intent id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_get_command(
                self.build_uri(UriTemplates.INTENTION_QUESTIONS, id)
            )
        )

    async def set_intent_questions_async(
        self,
        id: str,
        questions: list
    ) -> Command:
        """Create a intent question.

        Args:
            id (str): Intent id
            questions (list): Question

        Returns:
            Command: Command response
        """
        intent_questions_resource = self.__build_resource(
            questions,
            ContentType.QUESTION
        )

        return await self.process_command_async(
            self.create_set_command(
                self.build_uri(UriTemplates.INTENTION_QUESTIONS, id),
                intent_questions_resource,
                ContentTypes.COLLECTION
            )
        )

    async def delete_intent_question_async(
        self,
        id: str,
        question_id: str
    ) -> Command:
        """Delete a intent question.

        Args:
            id (str): Intent id
            question_id (str): Question id

        Returns:
            Command: Command response
        """
        return await self.process_command_async(
            self.create_delete_command(
                self.build_uri(
                    UriTemplates.INTENTION_QUESTION,
                    id,
                    question_id
                )
            )
        )

    def __build_resource(self, items: list, item_type: str) -> dict:
        return {
            'itemType': item_type,
            'items': items
        }



from typing import Callable

from lime_python import (ExternalAuthentication, KeyAuthentication,
                         PlainAuthentication, Transport)

from .application import Application
from .client import Client


class ClientBuilder:
    """Builder class to generate a SDK client."""

    def __init__(self) -> None:
        self.__application = Application()
        self.__transport_factory: Callable[[], Transport] = None

    def with_application(self, application: Application):
        self.__application = application
        return self

    def with_identifier(self, identifier: str):
        self.__application.identifier = identifier
        return self

    def with_instance(self, instance: str):
        self.__application.instance = instance
        return self

    def with_domain(self, domain: str):
        self.__application.domain = domain
        return self

    def with_scheme(self, scheme: str):
        self.__application.scheme = scheme
        return self

    def with_hostname(self, hostname: str):
        self.__application.hostname = hostname
        return self

    def with_port(self, port: int):
        self.__application.port = port
        return self

    def with_access_key(self, access_key: str):
        self.__application.authentication = KeyAuthentication(access_key)
        return self

    def with_password(self, password: str):
        self.__application.authentication = PlainAuthentication(password)
        return self

    def with_token(self, token: str, issuer: str = None):
        self.__application.authentication = ExternalAuthentication(
            token,
            issuer
        )
        return self

    def with_issuer(self, issuer: str):
        if isinstance(self.__application.authentication, ExternalAuthentication):  # noqa: E501
            self.__application.authentication.issuer = issuer
        else:
            self.__application.authentication = ExternalAuthentication(
                None, issuer
            )
        return self

    def with_compression(self, compression: str):
        self.__application.compression = compression
        return self

    def with_encryption(self, encryption: str):
        self.__application.encryption = encryption
        return self

    def with_routing_rule(self, routing_rule: str):
        self.__application.presence['routingRule'] = routing_rule
        return self

    def with_echo(self, echo: str):
        self.__application.presence['echo'] = echo
        return self

    def with_priority(self, priority: str):
        self.__application.presence['priority'] = priority
        return self

    def with_round_robin(self, round_robin: str):
        self.__application.presence['roundRobin'] = round_robin
        return self

    def with_notify_consumed(self, notify_consumed: bool):
        self.__application.notify_consumed = notify_consumed
        return self

    def with_transport_factory(
        self,
        transport_factory: Callable[[], Transport]
    ):
        self.__transport_factory = transport_factory
        return self

    def with_command_timeout(self, command_timeout: int):
        self.__application.command_timeout = command_timeout
        return self

    def build(self) -> Client:
        if self.__transport_factory is None:
            raise ValueError(
                'You must pass a Transport Factory using with_transport_factory before call build'  # noqa: E501
            )
        uri = f'{self.__application.scheme}://{self.__application.hostname}:{self.__application.port}'  # noqa: E501, WPS221
        return Client(uri, self.__transport_factory, self.__application)



from asyncio import Future, ensure_future, get_event_loop
from time import sleep
from typing import Any, Callable, Dict, List, Type

from lime_python import (ClientChannel, Command, CommandMethod, Envelope,
                         GuestAuthentication, KeyAuthentication, Message,
                         Notification, NotificationEvent, PlainAuthentication,
                         Reason, ReasonCode, Session, SessionState, Transport)

from .application import Application
from .extensions import (AiExtension, ChatExtension, ExtensionBase,
                         MediaExtension)
from .receiver import Receiver
from .utilities import ClassUtilities

MAX_CONNECTION_TRY_COUNT = 10


class Client:
    """blip API Client."""

    def __init__(
        self,
        uri: str,
        transport_factory: Callable[[], Transport],
        application: Application
    ) -> None:

        default_application = Application()
        self.application: Application = ClassUtilities.merge_dataclasses(
            default_application,
            application
        )

        self.session_future: Future = None
        self.__message_receivers: List[Receiver] = []
        self.__notification_receivers: List[Receiver] = []
        self.__command_receivers: List[Receiver] = []
        self.__command_resolves: Dict[str, Callable] = {}
        self.__session_finished_handlers: List[Callable[[Session], None]] = []
        self.__session_failed_handlers: List[Callable[[Session], None]] = []

        self.uri: str = uri
        self.__listening: bool = False
        self.__closing: bool = False
        self.__connection_try_count: int = 0

        if isinstance(transport_factory, Transport):
            transport_factory = lambda: transport_factory

        self.transport: Transport = transport_factory()
        self.__transport_factory: Callable[[], Transport] = transport_factory

        self.client_channel: ClientChannel = None

        self.__extensions: Dict[Type, ExtensionBase] = {}

        self.__initialize_client_channel()

    @property
    def chat_extension(self) -> ChatExtension:  # noqa: D102
        return self.__get_extension(ChatExtension)

    @property
    def media_extension(self) -> MediaExtension:  # noqa: D102
        return self.__get_extension(MediaExtension, self.application.domain)

    @property
    def ai_extension(self) -> AiExtension:  # noqa: D102
        return self.__get_extension(AiExtension, self.application.domain)

    @property
    def listening(self) -> bool:  # noqa: D102
        return self.__listening

    @listening.setter
    def listening(self, value: bool):
        self.__listening = value
        self.on_listening_changed(value)

    async def connect_with_guest_async(self, identifier: str) -> Session:
        """Connect using Guest Authentication.

        Args:
            identifier (str): the identifier

        Raises:
            ValueError: if identifier is not passed

        Returns:
            Session: the connected session
        """
        if not identifier:
            raise ValueError('identifier is required')
        self.application.identifier = identifier
        self.application.authentication = GuestAuthentication()
        return await self.connect_async()

    async def connect_with_password_async(
        self,
        identifier: str,
        password: str,
        presence: Dict[str, str] = None
    ) -> Session:
        """Connect using Plain Authentication.

        Args:
            identifier (str): the identifier
            password (str): the password
            presence (Dict[str, str]): presence to be sent.

        Raises:
            ValueError: if identifier or password is not passed

        Returns:
            Session: the connected session
        """
        if not identifier:
            raise ValueError('identifier is required')
        if not password:
            raise ValueError('password is required')

        self.application.identifier = identifier
        self.application.authentication = PlainAuthentication(password)

        if presence:
            self.application.presence = presence
        return await self.connect_async()

    async def connect_with_key_async(
        self,
        identifier: str,
        key: str,
        presence: Dict[str, str] = None
    ) -> Session:
        """Connect using Key Authentication.

        Args:
            identifier (str): the identifier
            key (str): the key
            presence (Dict[str, str]): presence to be sent.

        Raises:
            ValueError: if identifier or key is not passed

        Returns:
            Session: the connected session
        """
        if not identifier:
            raise ValueError('identifier is required')
        if not key:
            raise ValueError('key is required')

        self.application.identifier = identifier
        self.application.authentication = KeyAuthentication(key)

        if presence:
            self.application.presence = presence
        return await self.connect_async()

    async def connect_async(self) -> Session:
        """Open a connection on transport and start application.

        Raises:
            ConnectionError: connection tries exceeded.

        Returns:
            Session: The connected Session
        """
        if self.__connection_try_count >= MAX_CONNECTION_TRY_COUNT:
            raise ConnectionError(
                f'Could not connect: Max connection try count of \
                {MAX_CONNECTION_TRY_COUNT} reached. Please check you network \
                and refresh the page.')

        self.__connection_try_count += 1
        self.__closing = False

        await self.transport.open_async(self.uri)
        session = await self.client_channel.establish_session_async(
            self.application.compression,
            self.application.encryption,
            f'{self.application.identifier}@{self.application.domain}',
            self.application.authentication,
            self.application.instance
        )
        await self.__send_presence_command_async()
        await self.__send_receipts_command_async()

        self.listening = True
        self.__connection_try_count = 0

        return session

    async def close_async(self) -> Session:
        """Close the open connection.

        Returns:
            Session: the closed session
        """
        self.__closing = True

        if self.client_channel.state == SessionState.ESTABLISHED:
            return await self.client_channel.send_finishing_session_async()

        if self.session_future.done():
            return self.session_future.result()
        return await self.session_future

    def connect(self) -> Session:
        """Open a connection on transport and start application.

        Raises:
            ConnectionError: connection tries exceeded.

        Returns:
            Session: The connected Session
        """  # noqa: DAR402
        loop = get_event_loop()
        return loop.run_until_complete(self.connect_async())

    def close(self) -> Session:
        """Close the open connection.

        Returns:
            Session: the closed session
        """
        loop = get_event_loop()
        return loop.run_until_complete(self.close_async())

    def send_message(self, message: Message) -> None:
        """Send a Message.

        Args:
            message (Message): Message to be sent
        """
        self.client_channel.send_message(message)

    def send_notification(self, notification: Notification) -> None:
        """Send a Notification.

        Args:
            notification (Notification): Notification to be sent
        """
        self.client_channel.send_notification(notification)

    def send_command(self, command: Command) -> None:
        """Send a command.

        Args:
            command (Command): Command to be sent
        """
        self.client_channel.send_command(command)

    async def process_command_async(
        self,
        command: Command,
        timeout: float = None
    ) -> Command:
        """Process a Command asynchronously and return the result.

        Args:
            command (Command): The Command to be processed
            timeout (float): Timeout to process the Command

        Returns:
            Command: The result Command
        """
        timeout = timeout if timeout else self.application.command_timeout
        return await self.client_channel.process_command_async(
            command,
            timeout
        )

    def process_command(
        self,
        command: Command,
        timeout: float = None
    ) -> Command:
        """Process a Command and return the result.

        Args:
            command (Command): The Command to be processed
            timeout (float): Timeout to process the Command

        Returns:
            Command: The result Command
        """
        loop = get_event_loop()
        return loop.run_until_complete(
            self.process_command_async(command, timeout)
        )

    def add_message_receiver(self, receiver: Receiver) -> Callable[[], None]:
        """Add a message receiver.

        Args:
            receiver (Receiver): the Receiver

        Returns:
            Callable[[], None]: a method to remove the receiver
        """
        return self.__add_receiver(self.__message_receivers, receiver)

    def clear_message_receivers(self) -> None:
        """Remove all message receivers."""
        self.__message_receivers = []

    def add_command_receiver(self, receiver: Receiver) -> Callable[[], None]:
        """Add a command receiver.

        Args:
            receiver (Receiver): the Receiver

        Returns:
            Callable[[], None]: a method to remove the receiver
        """
        return self.__add_receiver(self.__command_receivers, receiver)

    def clear_command_receivers(self) -> None:
        """Remove all command receivers."""
        self.__command_receivers = []

    def add_notification_receiver(
        self,
        receiver: Receiver
    ) -> Callable[[], None]:
        """Add a notification receiver.

        Args:
            receiver (Receiver): the Receiver

        Returns:
            Callable[[], None]: a method to remove the receiver
        """
        return self.__add_receiver(self.__notification_receivers, receiver)

    def clear_notification_receivers(self) -> None:
        """Remove all notification receivers."""
        self.__notification_receivers = []

    def add_session_finished_handler(
        self,
        handler: Callable[[Session], None]
    ) -> Callable[[], None]:
        """Add a session finished handler.

        Args:
            handler (Callable[[Session], None]): the handler callback

        Returns:
            Callable[[], None]: a method to remove the handler
        """
        return self.__add_handler(self.__session_finished_handlers, handler)

    def clear_session_finished_handlers(self) -> None:
        """Remove all session finished handlers."""
        self.__session_finished_handlers = []

    def add_session_failed_handler(
        self,
        handler: Callable[[Session], None]
    ) -> Callable[[], None]:
        """Add a session failed handler.

        Args:
            handler (Callable[[Session], None]): the handler callback

        Returns:
            Callable[[], None]: a method to remove the handler
        """
        return self.__add_handler(self.__session_failed_handlers, handler)

    def clear_session_failed_handlers(self) -> None:
        """Remove all session failed handlers."""
        self.__session_failed_handlers = []

    def on_listening_changed(self, value: bool) -> None:
        """Handle callback to client listening changes.

        This method can be overwrited.

        Args:
            value (bool): the new listening value
        """
        pass

    def __initialize_client_channel(self) -> None:
        """Initialize client channel listeners."""
        self.transport.on_close = self.__transport_on_close

        self.session_future = Future()

        self.client_channel = ClientChannel(self.transport, True, False)
        self.client_channel.on_message = self.__client_channel_on_message
        self.client_channel.on_notification = self.__client_channel_on_notification  # noqa: E501
        self.client_channel.on_command = self.__client_channel_on_command
        self.client_channel.on_session_finished = self.__client_channel_on_session_finished  # noqa: E501
        self.client_channel.on_session_failed = self.__client_channel_on_session_failed  # noqa: E501

    def __add_handler(
        self,
        handler_list: List[Callable[[Session], None]],
        handler: Callable[[Session], None]
    ) -> Callable[[], None]:
        handler_list.append(handler)
        position = len(handler_list) - 1
        return lambda: handler_list.pop(position)

    def __add_receiver(
        self,
        receiver_list: List[Receiver],
        receiver: Receiver
    ) -> Callable[[], None]:
        receiver_list.append(receiver)
        return lambda: receiver_list.remove(receiver)

    async def __send_presence_command_async(self) -> Command:
        if isinstance(self.application.authentication, GuestAuthentication):
            return None

        command = Command(
            CommandMethod.SET,
            '/presence',
            'application/vnd.lime.presence+json',
            self.application.presence
        )
        return await self.process_command_async(command)

    async def __send_receipts_command_async(self) -> Command:
        if isinstance(self.application.authentication, GuestAuthentication):
            return None

        command = Command(
            CommandMethod.SET,
            '/receipt',
            'application/vnd.lime.receipt+json',
            {
                'events': [
                    NotificationEvent.FAILED,
                    NotificationEvent.ACCEPTED,
                    NotificationEvent.DISPATCHED,
                    NotificationEvent.RECEIVED,
                    NotificationEvent.CONSUMED
                ]
            }
        )
        return await self.process_command_async(command)

    def __client_channel_on_session_finished(self, session: Session) -> None:
        self.session_future.set_result(session)
        self.__notify_handlers(self.__session_finished_handlers, session)

    def __client_channel_on_session_failed(self, session: Session) -> None:
        self.session_future.set_exception(session)
        self.__notify_handlers(self.__session_failed_handlers, session)

    def __notify_handlers(
        self,
        handler_list: List[Callable[[Envelope], None]],
        envelope: Envelope
    ) -> None:
        for handler in handler_list:
            handler(envelope)

    def __client_channel_on_command(
        self,
        command: Command
    ) -> None:
        resolve = self.__command_resolves.get(command.id)
        resolve = resolve if resolve else self.__reflect
        resolve(command)

        self.__notify_receivers(self.__command_receivers, command)

    def __client_channel_on_notification(
        self,
        notification: Notification
    ) -> None:
        self.__notify_receivers(self.__notification_receivers, notification)

    def __notify_receivers(
        self,
        receiver_list: List[Receiver],
        envelope: Envelope
    ) -> None:
        [   # noqa: WPS428
            receiver.callback(envelope)
            for receiver in receiver_list
            if receiver.predicate(envelope)
        ]

    def __client_channel_on_message(self, message: Message) -> None:
        should_notify = message.id and (
            not message.to or
            self.client_channel.local_node.lower().startswith(
                message.to.lower()
            )
        )

        if should_notify:
            notification = Notification(NotificationEvent.RECEIVED)
            notification.id = message.id
            notification.to = message.pp if message.pp else message.from_n
            notification.metadata = {
                '#message.to': message.to
            }
            self.send_notification(notification)

        try:
            self.__notify_message_receivers(message)
            self.__notify(should_notify, message)
        except Exception as error:
            self.__notify(should_notify, message, error)

    def __transport_on_close(self) -> None:
        self.listening = False
        if not self.__closing:
            # Use an exponential backoff for the timeout in seconds
            timeout = (100 * (2 ** self.__connection_try_count)) / 1000.0

            # Try to reconnect after the timeout
            sleep(timeout)
            if not self.__closing:
                self.transport = self.__transport_factory()
                self.__initialize_client_channel()
                ensure_future(self.connect_async())

    def __notify_message_receivers(self, message: Message) -> None:
        for receiver in self.__message_receivers:
            result: bool = None
            if receiver.predicate(message):
                result = receiver.callback(message)
            if result is False:
                raise ValueError

    def __notify(
        self,
        should_notify: bool,
        message: Message,
        error: Exception = None
    ) -> None:
        if not should_notify:
            return

        if error:
            notification = Notification(
                NotificationEvent.FAILED,
                Reason(
                    ReasonCode.APPLICATION_ERROR,
                    str(error)
                )
            )
            notification.id = message.id
            notification.from_n = message.from_n
            self.send_notification(notification)

        if self.application.notify_consumed:
            notification = Notification(NotificationEvent.CONSUMED)
            notification.id = message.id
            notification.to = message.pp if message.pp else message.from_n
            notification.metadata = {
                '#message.to': message.to
            }
            self.send_notification(notification)

    def __get_extension(self, type: Type, to: str = None) -> ExtensionBase:
        extension = self.__extensions.get(type)
        if not extension:
            extension = type(self, to)
            self.__extensions[type] = extension
        return extension

    def __reflect(self, any: Any) -> Any:
        return any



class UriTemplates:
    """AI Model uri templates."""

    MODELS = '/models'

    MODEL = '/model/{0}'

    MODELS_SUMMARY = '/models/summary'

    LAST_TRAINED_OR_PUBLISH_MODEL = '/models/last-trained-or-published'



class ContentType:
    """Content Assistant Content type class."""

    CONTENT_RESULT = 'application/vnd.iris.ai.content-result+json'

    CONTENT_COMBINATION = 'application/vnd.iris.ai.content-combination+json'

    ANALYSIS = 'application/vnd.iris.ai.analysis-request+json'



class ContentType:
    """Intents content type."""

    INTENTION = 'application/vnd.iris.ai.intention+json'

    ANSWER = 'application/vnd.iris.ai.answer+json'

    QUESTION = 'application/vnd.iris.ai.question+json'



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict
from urllib.parse import urlencode
from uuid import uuid4

from lime_python import Command, CommandMethod

from ..utilities import RequestUtilities

if TYPE_CHECKING:
    from ..client import Client


class ExtensionBase:
    """Class base to all sdk extensions."""

    def __init__(self, client: Client, to: str = None) -> None:
        self.client = client
        self.to = to

    def create_get_command(
        self,
        uri: str,
        id: str = None
    ) -> Command:
        """Create a get Command.

        Args:
            uri (str): Command uri
            id (str): Comand id

        Returns:
            Command
        """
        command = Command(CommandMethod.GET, uri)
        command.id = id if id else str(uuid4())

        if self.to:
            command.to = self.to

        return command

    def create_set_command(
        self,
        uri: str,
        resource: Any,
        type_n: str = None,
        id: str = None
    ) -> Command:
        """Create a set Command.

        Args:
            uri (str): Command uri
            type_n (str): resource mime type
            resource (Any): Command resource
            id (str): Command id

        Returns:
            Command
        """
        command = Command(CommandMethod.SET, uri, type_n, resource)
        command.id = id if id else str(uuid4())

        if self.to:
            command.to = self.to

        return command

    def create_merge_command(
        self,
        uri: str,
        resource: Any,
        type_n: str = None,
        id: str = None
    ) -> Command:
        """Create a merge Command.

        Args:
            uri (str): Command uri
            type_n (str): resource mime type
            resource (Any): Command resource
            id (str): Command id

        Returns:
            Command
        """
        command = Command(CommandMethod.MERGE, uri, type_n, resource)
        command.id = id if id else str(uuid4())

        if self.to:
            command.to = self.to

        return command

    def create_delete_command(
        self,
        uri: str,
        id: str = None
    ) -> Command:
        """Create a delete Command.

        Args:
            uri (str): Command uri
            id (str): Command id

        Returns:
            Command
        """
        command = Command(CommandMethod.DELETE, uri)
        command.id = id if id else str(uuid4())

        if self.to:
            command.to = self.to

        return command

    async def process_command_async(
        self,
        command: Command
    ) -> Command:
        """Process a command async.

        Args:
            command (Command): the Command to process

        Returns:
            Command: the response
        """
        command.id = command.id if command.id else str(uuid4())
        return await self.client.process_command_async(command)

    def build_resource_query(
        self,
        uri: str,
        query: Dict[str, str]
    ) -> str:
        """Build the resource query.

        Usage:

        ```py
        build_resource_query('https://take.net', { 'foo': 'bar' })
        # output: 'https://take.net?foo=bar'
        build_uri('https://take.net?', { 'foo': 'bar', 'zoo': 'buu' })
        # output: 'https://take.net?foo=bar&zoo=buu'
        ```

        Args:
            uri (str): base uri
            query (Dict[str, str]): items to add

        Returns:
            str: final uri
        """
        fixed_query = query.copy()
        for name, value in query.items():
            if value is None:
                del fixed_query[name]

        if len(fixed_query) and not uri.endswith('?'):
            uri += '?'  # noqa: WPS336

        encoded_part = urlencode(fixed_query, quote_via=RequestUtilities.quote)

        return f'{uri}{encoded_part}'

    def build_uri(self, uri: str, *args: dict) -> str:
        """Build a uri with parameters.

        Usage:

        ```py
        build_uri('https://take.net/{1}', 'foo')
        # output: 'https://take.net/foo'
        build_uri('https://take.net/{1}/{2}/{3}', 'foo', 'bar', 'zoo')
        # output: 'https://take.net/foo/bar/zoo'
        ```

        Args:
            uri (str): the template uri with {index}
            args: the parameters to replace

        Returns:
            str: the final uri
        """
        for index, value in enumerate(args):
            uri = uri.replace(f'{{{index}}}', RequestUtilities.quote(value))
        return uri
